/*
  buffalo.c - Buffalo specific feature
  Copyright (C) BUFFALO INC.
 */

#include "common.h"

#if defined(BUFFALO_FEATURE)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <ctype.h>

#include "buffalo.h"

#define SLAVE_SHUTDOWN_DEFAULT	"/usr/local/sbin/slave_shutdown"
#define ERROR_NOTIFY_SCRIPT	"/usr/local/bin/lcd_error_man.sh"
#define BUFFALO_ERROR_STATUS	"/tmp/error_status"
#define BUFFALO_UPS_STATUS	"/tmp/ups_status"
#define BUFFALO_BOOTING		"/proc/buffalo/booting"
#define BUFFALO_LIBBUFFALO_BIN	"/usr/local/bin/libbuffalo_bin"
#define BUFFALO_UPS_INFO	"/etc/melco/ups"
static char *slave_shutdown_cmd = SLAVE_SHUTDOWN_DEFAULT;

#define SLAVE_LIST_DEFAULT	"legacy_slave"
static char *slave_list_path = NULL;
static int last_status_is_ob[2] = {};
static int last_status_is_nc[2] = {};

#define ETC_NAS_FEATURE			"/etc/nas_feature"
#define SUPPORT_ISCSI_TYPE		"SUPPORT_ISCSI_TYPE"
#define SUPPORT_ISCSI_TYPE_RTS		"RTS"
#define RTS_ISCSI_GET_ISCONNECTED_PY	"/usr/local/lib/iscsi/iscsi_get_IsConnected.py"
#define RTS_ISCSI_GET_ISCONNECTED_PYC	"/usr/local/lib/iscsi/iscsi_get_IsConnected.pyc"

static int GetNasFeatureVal(const char *key, char **data);

static int
get_result_system(char *cmd)
{
	int status;
	int result;

	status = system(cmd);
	if( status != -1 ) {
		result = WEXITSTATUS(status);
	}else{
		result = status;
	}

	return( result );
}

static int
get_result_cmd(char *cmd)
{
	FILE *fp;
	int result = 0;

	fp = popen(cmd, "r");
	(void)fscanf(fp, "%d", &result);
	pclose(fp);

	return( result );
}

static void
remove_space(char *key)
{
	while(*key) {
		if( isspace(*key) ) {
			*key=0;
		}
		key++;
	}
}

static void
remove_last_crlf(char *str)
{
	int len=strlen(str);
	while(len--) {
		if( str[len]=='\n'||str[len]=='\r' ) {
			str[len]=0;
		}else{
			break;
		}
	}
}

static char *
getNextKey(const char *buf)
{
	int len;
	len=strlen(buf);
	// skip space
	while(len-- && *buf!='\r' && *buf!='\n') {
		if( !(*buf==' '||*buf=='\t') ) {
			return (char *)buf;
		}
		buf++;
	}
	return( (char *)buf );
}

static int
IsFileExist(const char *path)
{
	struct stat sb;
	FILE *fp = NULL;

	fp = fopen(path, "r");
	if( !fp ) {
		return(0);
	}
	fclose(fp);

	if( stat(path, &sb) != 0 ) {
		return(0);
	}

	if( S_ISREG(sb.st_mode) ) {
		return(1);
	}

	return(0);
}

static char *
get_StringbyKey(const char *const fname, const char *const req_key, char *dst, int dstlen)
{
	FILE *fp;
	char buffer[512];

	if( !IsFileExist(fname) ) {
		return NULL;
	}

	if( (fp=fopen(fname,"r")) == NULL ) return(NULL);

	while( fgets(buffer,sizeof(buffer),fp) != NULL ) {
		char *key,*value;

		key   = getNextKey(buffer);
		value = strchr(key,'=');

		if( key[0]=='#' || value==NULL ) {
			continue;
		}

		value[0] = '\0';
		value ++;

		remove_space(key);
		remove_last_crlf(value);

		if( strcmp(key,req_key)==0 ) {
			strncpy(dst,value,dstlen-1);
			fclose(fp);
			return(dst);
		}
	}

	fclose(fp);
	return(NULL);
}

void
buffalo_set_slave_shutdown_cmd(char *cmd)
{
	char *new_cmd;

	if (cmd == NULL) {
		return;
	}

	new_cmd = (char *)malloc(strlen(cmd) + 1);
	if (new_cmd == NULL) {
		upslogx(LOG_ERR, "malloc failed!");
		return;
	}
	memset(new_cmd, 0, sizeof(*new_cmd));
	strcpy(new_cmd, cmd);

	slave_shutdown_cmd = new_cmd;

	return;
}

void
buffalo_set_default_slave_list_path(void)
{
	char *new_path;
	char buf[1024];

	memset(buf, 0, sizeof(buf));
	sprintf(buf, "%s/%s", confpath(), SLAVE_LIST_DEFAULT);
	new_path = (char *)malloc(strlen(buf) + 1);
	if (new_path == NULL) {
		upslogx(LOG_ERR, "malloc failed!");
		return;
	}
	memset(new_path, 0, sizeof(*new_path));
	strcpy(new_path, buf);

	slave_list_path = new_path;

	return;
}

void
buffalo_set_slave_list_path(char *list)
{
	char *new_path;

	if (list == NULL) {
		return;
	}

	new_path = (char *)malloc(strlen(list) + 1);
	if (new_path == NULL) {
		upslogx(LOG_ERR, "malloc failed!");
		return;
	}
	memset(new_path, 0, sizeof(*new_path));
	strcpy(new_path, list);

	slave_list_path = new_path;

	return;
}

int
buffalo_shutdown_legacy_slave(void)
{
	int ret;
	char ext_cmd[4096];

	memset(ext_cmd, 0, sizeof(ext_cmd));
	ret = sprintf(ext_cmd, "%s %s", slave_shutdown_cmd, slave_list_path);

	upslogx(LOG_INFO, "execute external command %s", ext_cmd);
	ret = system(ext_cmd);

	return ret;
}

static int
buffalo_write_out_ups_status(char *name, const char *statword, int isamaster)
{
	FILE *fp = NULL;
	char fn[256];

	upsdebugx(1, "%s : %s statword = %s\n", __FUNCTION__, name, statword);

	if(isamaster || (strcmp(statword, "FSD")))
	{
		snprintf(fn, sizeof(fn), "/tmp/%s_status", name);
		upsdebugx(1, "%s : %s updating %s\n", __FUNCTION__, name, fn);
		fp = fopen(fn, "w");
		if(fp == NULL)
			return -1;

		if(strstr(statword, "OL"))
		{
			upsdebugx(1, "%s : %s ups_status=0\n", name, __FUNCTION__);
			fprintf(fp, "ups_status=0\n");
		}
		else if(strstr(statword, "OB"))
		{
			upsdebugx(1, "%s : %s ups_status=1\n", name, __FUNCTION__);
			fprintf(fp, "ups_status=1\n");
		}
		else if(strstr(statword, "LB"))
		{
			upsdebugx(1, "%s : %s ups_status=1\n", name, __FUNCTION__);
			fprintf(fp, "ups_status=1\n");
		}
		else
		{
			upsdebugx(1, "%s : %s ups_status=2\n", name, __FUNCTION__);
			fprintf(fp, "ups_status=2\n");
		}

		if(strstr(statword, "LB"))
		{
			upsdebugx(1, "%s : %s battery_low=1\n", name, __FUNCTION__);
			fprintf(fp, "battery_low=1\n");
		}
		else
		{
			upsdebugx(1, "%s : %s battery_low=0\n", name, __FUNCTION__);
			fprintf(fp, "battery_low=0\n");
		}

		if(strstr(statword, "FSD"))
		{
			upsdebugx(1, "%s : %s shutdown_active=waiting\n", name, __FUNCTION__);
			fprintf(fp, "shutdown_active=waiting\n");
		}
		else
		{
			upsdebugx(1, "%s : %s shutdown_active=normal\n", name, __FUNCTION__);
			fprintf(fp, "shutdown_active=normal\n");
		}

		fclose(fp);
	}
	else
	{

	}
	return 0;
}

int
buffalo_notify_error(char *name, const char *statword, int isamaster, const char *inputvol)	//	"add inputvol status"
{
	char cmd[1024];
	double	inputv;	//	change buffer for input.voltage
	FILE *fp;
	char buf[128];
	char *ups_num = NULL;
	int i = 1;
	char err_name[16] = {0};

	upsdebugx(1, "%s: name:%s statword:%s master:%d", __FUNCTION__, name, statword, isamaster);
	strcpy(err_name, name);
	if((strcmp(name, "ups") == 0) || (strcmp(name, "ups1") == 0)) {
		i = 0;
		if(strcmp(name, "ups1") == 0){
			GetNasFeatureVal(NF_KEY_MULTI_UPS_MONITORING_N, &ups_num);
			if (ups_num == NULL || atoi(ups_num) == 1 || strcmp(ups_num, "") == 0 || strcmp(ups_num, "\0") == 0){
				memset(err_name, 0, 6);
				strcpy(err_name, "ups");
			}
			if (ups_num != NULL) {
				free(ups_num);
			}
		}
	}

	if(strstr(statword, "OB"))
	{
		inputv = atof(inputvol);
		if( inputv <= 30.0 )
		{
			upslogx(LOG_INFO, "Not UPS TEST_mode. UPS status on battery");
			if(last_status_is_ob[i] == 0)
			{
				last_status_is_ob[i] = 1;
				memset(cmd, 0, sizeof(cmd));
				sprintf(cmd, "%s %s_error on diag_on buzzer_on", ERROR_NOTIFY_SCRIPT, err_name);
				upsdebugx(1, "%s: %s", __FUNCTION__, cmd);
				system(cmd);
			}
		}
		else
		{
			upslogx(LOG_INFO, "UPS TEST_mode");
		}
	}
	else if(strstr(statword, "OL"))
	{
		fp = fopen(BUFFALO_ERROR_STATUS, "r");
		if(fp != NULL){
			while(fgets(buf,sizeof(buf),fp) != NULL ) {
				if(strstr(buf, "ups")){
					last_status_is_ob[i] = 0;
					last_status_is_nc[i] = 0;
					memset(cmd, 0, sizeof(cmd));
					sprintf(cmd, "%s %s_error off", ERROR_NOTIFY_SCRIPT, err_name);
					upsdebugx(1, "%s: %s", __FUNCTION__, cmd);
					system(cmd);
				}
			}
			fclose(fp);
		}
	}
	else if(strstr(statword, "NC"))
	{
		if(last_status_is_nc[i] == 0)
		{
			last_status_is_nc[i] = 1;
			memset(cmd, 0, sizeof(cmd));
			sprintf(cmd, "%s %s_error on diag_on buzzer_on", ERROR_NOTIFY_SCRIPT, err_name);
			system(cmd);
		}
	}

	return 0;
}

void
buffalo_do_misc(char *name, const char *statword, int isamaster, const char *inputvol)
{
//	upslogx(LOG_INFO, "%s : statword = %s\n", __FUNCTION__, statword);
	buffalo_write_out_ups_status(name, statword, isamaster);
	buffalo_notify_error(name, statword, isamaster, inputvol);
}

int
buffalo_is_booting(void)
{
	FILE *fp = NULL;
	char buf[1024];

	fp = fopen(BUFFALO_BOOTING, "r");
	if(fp == NULL)
		return 0;

	memset(buf, 0, sizeof(buf));
	fgets(buf, sizeof(buf), fp);
	fclose(fp);

	if(strncmp(buf, "1", 1) == 0)
		return 1;
	return 0;
}

int
buffalo_is_iscsi_session_exist(void)
{
	char cmd[1024];
	char val[256];
	int is_session = 0;
	memset(cmd, 0, sizeof(cmd));
	memset(val, 0, sizeof(val));

	/* check SUPPORT_ISCSI_TYPE */
	if( NULL != get_StringbyKey(ETC_NAS_FEATURE, SUPPORT_ISCSI_TYPE, val, sizeof(val)) ) {
		if( !strcmp(val, SUPPORT_ISCSI_TYPE_RTS) ) {
			/* check rts-api file is exists */
			if( IsFileExist(RTS_ISCSI_GET_ISCONNECTED_PYC) ) {
				sprintf(cmd, "%s", RTS_ISCSI_GET_ISCONNECTED_PYC);
			}else if( IsFileExist(RTS_ISCSI_GET_ISCONNECTED_PY) ) {
				sprintf(cmd, "%s", RTS_ISCSI_GET_ISCONNECTED_PY);
			}else{
				upslogx(LOG_ERR, "file does not exists [%s]", RTS_ISCSI_GET_ISCONNECTED_PY);
				return is_session;
			}
			is_session = get_result_cmd(cmd);
		}else{
			sprintf(cmd, "%s iSCSI IsConnected", BUFFALO_LIBBUFFALO_BIN);
			is_session = get_result_system(cmd);
		}
	}else{
		sprintf(cmd, "%s iSCSI IsConnected", BUFFALO_LIBBUFFALO_BIN);
		is_session = get_result_system(cmd);
	}
	upslogx(LOG_INFO, "iscsi(%s) connection [%d]", val, is_session);
	return is_session;
}

int
buffalo_get_ups_wait_until_disconnect(char *upsname)
{
	FILE *fp = NULL;
	char line[1024];
	char var[64];
	int is_iscsi_wait = 0, var_len = 0;

	if (upsname == NULL)
		return 0;


	fp = fopen(BUFFALO_UPS_INFO, "r");
	if (fp == NULL)
		return 0;

	snprintf(var, sizeof(var), "%s_wait_until_disconnect", upsname);
	var[sizeof(var) - 1] = '\0';
	var_len = strlen(var);

	memset(line, 0, sizeof(line));
	while (fgets(line, sizeof(line), fp)) {
		if (strncmp(line, var, var_len) == 0) {
			if (strncmp((const char*)&line[var_len+1], "on", 2) == 0) {
				is_iscsi_wait = 1;
				break;
			}
		}
	}
	upslogx(LOG_INFO, "%s_wait_until_disconnect=%d", upsname, is_iscsi_wait);
	fclose(fp);

	if (is_iscsi_wait == 1) {
		return 1;
	}
	return 0;
}

int buffalo_ups_failed_by_psu_ac_module(char *upsname)
{
	int rc = 0;
	char cmd[64];
	snprintf(cmd, sizeof(cmd), "/usr/local/bin/ups_failed_by_psu.sh %s", upsname);
	int system_rc = 0;
	system_rc = system(cmd);
	if (system_rc != -1) {
		system_rc = WEXITSTATUS(system_rc);
	}
	upslogx(LOG_DEBUG, "%s (os_rc:%d)", __func__, cmd, system_rc);
	if (0 == system_rc) {
		upslogx(LOG_ERR, "%s: AC module failed for %s (rc:%d)", __func__, upsname, system_rc);
		rc = 1;
	}
	return rc;
}


static int
GetNasFeatureVal(const char *key, char **data)
{
	int	ret = -1;
	char	cmd[128];
	FILE	*fp;
	char	tmp[256];
	char	*p;

	errno	= 0;
	fp	= NULL;

	if (*data != NULL) {
		errno	= EINVAL;
		return ret;
	}

	*data = (char *)calloc(sizeof(char), sizeof(tmp));
	if (*data == NULL) {
		fprintf(
			stderr,
			"%s(%d): malloc failed.\n",
			__FUNCTION__,
			__LINE__
			);
		return ret;
	}

	memset(tmp, 0, sizeof(tmp));
	memset(cmd, 0, sizeof(cmd));
	snprintf(cmd, sizeof(cmd), CMD_LIBBUFFALO_BIN_READ_MAX_UPS, key);

	if ((fp = popen(cmd, "r")) == NULL) {
		return ret;
	}

	while (fgets(tmp, sizeof(tmp), fp) != NULL);
	ret = pclose(fp);

	if (ret == -1 && errno == ECHILD)
		ret = 0;

	if (ret == 0) {
		p = strrchr(tmp, '\n');

		if (p != NULL) {
			*p = '\0';
		}

		p = *data;
		strncpy(p, tmp, sizeof(tmp) - 1);
		*(p + (sizeof(tmp) - 1)) = '\0';
	}
	return ret;
}

#endif	/* BUFFALO_FEATURE */
